from abc import ABC, abstractmethod
from typing import AsyncGenerator

from openai.types.chat import ChatCompletionMessage

from ..types import Agent
from ..util import run_sync


class BaseClient(ABC):
    @abstractmethod
    async def aget_chat_completion(
        self,
        agent: Agent,
        history: list,
        context_variables: dict,
        model_override: str,
        stream: bool,
        debug: bool,
    ) -> ChatCompletionMessage | AsyncGenerator:
        pass

    def get_chat_completion(
        self,
        agent: Agent,
        history: list,
        context_variables: dict,
        model_override: str,
        stream: bool,
        debug: bool,
    ) -> ChatCompletionMessage:
        return run_sync(
            self.aget_chat_completion(
                agent=agent,
                history=history,
                context_variables=context_variables,
                model_override=model_override,
                stream=stream,
                debug=debug,
            )
        )



import copy
import json
from collections import defaultdict
from typing import AsyncGenerator

from .util import merge_chunk, run_sync, debug_print
from .clients import BaseClient, OpenAIClient
from .tools import handle_tool_calls
from .types import (
    Agent,
    Response,
    Function,
    ChatCompletionMessageToolCall,
)


class Troop:
    def __init__(self, client=None):
        self.client: BaseClient = OpenAIClient() if not client else client

    async def arun_and_stream(
        self,
        agent: Agent,
        messages: list,
        context_variables: dict = {},
        model_override: str = None,
        debug: bool = False,
        max_turns: int = float("inf"),
        execute_tools: bool = True,
    ):
        active_agent = agent
        context_variables = copy.deepcopy(context_variables)
        history = copy.deepcopy(messages)
        init_len = len(messages)

        while len(history) - init_len < max_turns:
            message = {
                "content": "",
                "sender": agent.name,
                "role": "assistant",
                "function_call": None,
                "tool_calls": defaultdict(
                    lambda: {
                        "function": {"arguments": "", "name": ""},
                        "id": "",
                        "type": "",
                    }
                ),
            }

            # get completion with current history, agent
            completion = await self.client.aget_chat_completion(
                agent=active_agent,
                history=history,
                context_variables=context_variables,
                model_override=model_override,
                stream=True,
                debug=debug,
            )

            yield {"delim": "start"}
            async for chunk in completion:
                delta = json.loads(chunk.choices[0].delta.model_dump_json())
                if delta["role"] == "assistant":
                    delta["sender"] = active_agent.name
                yield delta
                delta.pop("role", None)
                delta.pop("sender", None)
                merge_chunk(message, delta)
            yield {"delim": "end"}

            message["tool_calls"] = list(message.get("tool_calls", {}).values())
            if not message["tool_calls"]:
                message["tool_calls"] = None
            debug_print(debug, "Received completion:", message)
            history.append(message)

            if not message["tool_calls"] or not execute_tools:
                debug_print(debug, "Ending turn.")
                break

            # convert tool_calls to objects
            tool_calls = []
            for tool_call in message["tool_calls"]:
                function = Function(
                    arguments=tool_call["function"]["arguments"],
                    name=tool_call["function"]["name"],
                )
                tool_call_object = ChatCompletionMessageToolCall(
                    id=tool_call["id"], function=function, type=tool_call["type"]
                )
                tool_calls.append(tool_call_object)

            # handle function calls, updating context_variables, and switching agents
            partial_response = await handle_tool_calls(
                tool_calls, active_agent.functions, context_variables, debug
            )
            history.extend(partial_response.messages)
            context_variables.update(partial_response.context_variables)
            if partial_response.agent:
                active_agent = partial_response.agent

        yield {
            "response": Response(
                messages=history[init_len:],
                agent=active_agent,
                context_variables=context_variables,
            )
        }

    async def arun(
        self,
        agent: Agent,
        messages: list,
        context_variables: dict = {},
        model_override: str = None,
        stream: bool = False,
        debug: bool = False,
        max_turns: int = float("inf"),
        execute_tools: bool = True,
    ) -> Response | AsyncGenerator:
        """
        Run the agent asynchronously.
        
        Args:
            agent: The agent to run
            messages: The message history
            context_variables: Variables available to the agent
            model_override: Override the agent's model
            stream: If True, return an AsyncGenerator that yields chunks
            debug: Enable debug logging
            max_turns: Maximum number of turns
            execute_tools: Whether to execute tool calls
            
        Returns:
            Response if stream=False
            AsyncGenerator if stream=True
        """
        if stream:
            return self.arun_and_stream(
                agent=agent,
                messages=messages,
                context_variables=context_variables,
                model_override=model_override,
                debug=debug,
                max_turns=max_turns,
                execute_tools=execute_tools,
            )
        active_agent = agent
        context_variables = copy.deepcopy(context_variables)
        history = copy.deepcopy(messages)
        init_len = len(messages)

        while len(history) - init_len < max_turns and active_agent:
            # get completion with current history, agent
            completion = await self.client.aget_chat_completion(
                agent=active_agent,
                history=history,
                context_variables=context_variables,
                model_override=model_override,
                stream=stream,
                debug=debug,
            )
            message = completion.choices[0].message
            debug_print(debug, "Received completion:", message)
            message.sender = active_agent.name
            history.append(
                json.loads(message.model_dump_json())
            )  # to avoid OpenAI types (?)

            if not message.tool_calls or not execute_tools:
                debug_print(debug, "Ending turn.")
                break

            # handle function calls, updating context_variables, and switching agents
            partial_response = await handle_tool_calls(
                message.tool_calls, active_agent.functions, context_variables, debug
            )
            history.extend(partial_response.messages)
            context_variables.update(partial_response.context_variables)
            if partial_response.agent:
                active_agent = partial_response.agent

        return Response(
            messages=history[init_len:],
            agent=active_agent,
            context_variables=context_variables,
        )

    def run_and_stream(
        self,
        agent: Agent,
        messages: list,
        context_variables: dict = {},
        model_override: str = None,
        debug: bool = False,
        max_turns: int = float("inf"),
        execute_tools: bool = True,
    ):
        """Synchronous version of arun_and_stream"""
        async_gen = self.arun_and_stream(
            agent=agent,
            messages=messages,
            context_variables=context_variables,
            model_override=model_override,
            debug=debug,
            max_turns=max_turns,
            execute_tools=execute_tools,
        )

        # Convert async generator to sync generator
        while True:
            try:
                yield run_sync(async_gen.__anext__())
            except StopAsyncIteration:
                break

    def run(
        self,
        agent: Agent,
        messages: list,
        context_variables: dict = {},
        model_override: str = None,
        stream: bool = False,
        debug: bool = False,
        max_turns: int = float("inf"),
        execute_tools: bool = True,
    ) -> Response:
        """Synchronous version of arun"""
        if stream:
            return self.run_and_stream(
                agent=agent,
                messages=messages,
                context_variables=context_variables,
                model_override=model_override,
                debug=debug,
                max_turns=max_turns,
                execute_tools=execute_tools,
            )
        return run_sync(
            self.arun(
                agent=agent,
                messages=messages,
                context_variables=context_variables,
                model_override=model_override,
                stream=stream,
                debug=debug,
                max_turns=max_turns,
                execute_tools=execute_tools,
            )
        )



import json
from functools import wraps
from inspect import getdoc
from typing import TypeVar, get_type_hints

from openai import AsyncOpenAI
from pydantic import BaseModel

from .clients import OpenAIClient, BaseClient
from .types import Agent

T = TypeVar("T")


def ai_flow(client: BaseClient = OpenAIClient()):
    """
    A decorator that transforms a function into an AI-powered function using OpenAI's GPT.
    The function's docstring becomes the system prompt, its return type becomes the response format,
    and its parameters are converted into a JSON-like string for the user message.

    Example:
        @ai_function
        def summarize_text(text: str, max_words: int = 100) -> SummaryResponse:
            '''Generate a concise summary of the given text.'''
            ...
    """
    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):

            # Get the function's signature
            bound_args = func.__annotations__
            param_names = list(bound_args.keys())[:-1]  # Exclude return annotation

            params = {}
            if args:
                for i, arg in enumerate(args):
                    if i < len(param_names):
                        params[param_names[i]] = arg
            params.update(kwargs)

            user_content = json.dumps(params, indent=2)
            system_prompt = getdoc(func) or ""
            return_type = get_type_hints(func).get("return", str)
            # Always use JSON response format for BaseModel and dict return types
            response_format = (
                {"type": "json_object"}
                if issubclass(return_type, (dict, BaseModel))
                else None
            )

            # For Pydantic models, append the expected schema to the system prompt
            if issubclass(return_type, BaseModel):
                schema = return_type.model_json_schema()
                system_prompt = f"{system_prompt}\n\nResponse must be a JSON object matching this schema:\n{json.dumps(schema, indent=2)}"

            response = await client.aget_chat_completion(
                agent=Agent(
                    model="gpt-4o",
                    instructions=system_prompt,
                ),
                history=[{"role": "user", "content": user_content}],
                context_variables={},
                model_override=None,
                stream=False,
                debug=False,
            )
            result = response.choices[0].message.content

            if issubclass(return_type, BaseModel):
                return return_type.model_validate_json(result)
            elif issubclass(return_type, dict):
                return json.loads(result)
            else:
                return result

        return wrapper
    return decorator



import json
import inspect

from .types import AgentFunction, ChatCompletionMessageToolCall, Response, Result, Agent
from .util import debug_print

__CTX_VARS_NAME__ = "context_variables"


async def handle_tool_calls(
    tool_calls: list[ChatCompletionMessageToolCall],
    functions: list[AgentFunction],
    context_variables: dict,
    debug: bool,
) -> Response:
    """Handle tool calls and return a partial response with results."""
    function_map = {f.__name__: f for f in functions}
    partial_response = Response(messages=[], agent=None, context_variables={})

    for tool_call in tool_calls:
        name = tool_call.function.name
        # handle missing tool case, skip to next tool
        if name not in function_map:
            debug_print(debug, f"Tool {name} not found in function map.")
            partial_response.messages.append(
                {
                    "role": "tool",
                    "tool_call_id": tool_call.id,
                    "tool_name": name,
                    "content": f"Error: Tool {name} not found.",
                }
            )
            continue

        args = json.loads(tool_call.function.arguments)
        debug_print(debug, f"Processing tool call: {name} with arguments {args}")

        func = function_map[name]
        # pass context_variables to agent functions
        if __CTX_VARS_NAME__ in func.__code__.co_varnames:
            args[__CTX_VARS_NAME__] = context_variables

        raw_result = (
            await func(**args)
            if inspect.iscoroutinefunction(func)
            else func(**args)
        )

        result = Result.from_raw(raw_result)
        partial_response.messages.append(
            {
                "role": "tool",
                "tool_call_id": tool_call.id,
                "tool_name": name,
                "content": result.value,
            }
        )
        partial_response.context_variables.update(result.context_variables)
        if result.agent:
            partial_response.agent = result.agent

    return partial_response

def function_to_json(func) -> dict:
    """
    Converts a Python function into a JSON-serializable dictionary
    that describes the function's signature, including its name,
    description, and parameters.

    Args:
        func: The function to be converted.

    Returns:
        A dictionary representing the function's signature in JSON format.
    """
    type_map = {
        str: "string",
        int: "integer",
        float: "number",
        bool: "boolean",
        list: "array",
        dict: "object",
        type(None): "null",
    }

    try:
        signature = inspect.signature(func)
    except ValueError as e:
        raise ValueError(
            f"Failed to get signature for function {func.__name__}: {str(e)}"
        )

    parameters = {}
    for param in signature.parameters.values():
        try:
            param_type = type_map.get(param.annotation, "string")
        except KeyError as e:
            raise KeyError(
                f"Unknown type annotation {param.annotation} for parameter {param.name}: {str(e)}"
            )
        parameters[param.name] = {"type": param_type}

    required = [
        param.name
        for param in signature.parameters.values()
        if param.default == inspect._empty
    ]

    return {
        "type": "function",
        "function": {
            "name": func.__name__,
            "description": func.__doc__ or "",
            "parameters": {
                "type": "object",
                "properties": parameters,
                "required": required,
            },
        },
    }

def functions_to_tools(functions: list[AgentFunction]) -> list[dict]:
    """Returns a list of tool definitons based on the agent's functions."""
    tools_ = []
    for f in functions:
        tool = function_to_json(f)
        params = tool["function"]["parameters"]
        params["properties"].pop(__CTX_VARS_NAME__, None)
        if __CTX_VARS_NAME__ in params["required"]:
            params["required"].remove(__CTX_VARS_NAME__)
    return tools_ if tools_ else None

def handle_function_result(result, debug) -> Result:
    match result:
        case Result() as result:
            return result

        case Agent() as agent:
            return Result(
                value=json.dumps({"assistant": agent.name}),
                agent=agent,
            )
        case _:
            try:
                str_result = str(result)  # Try conversion first
                return Result(value=str_result)
            except Exception as e:
                error_message = f"Failed to cast response to string: {result}. Make sure agent functions return a string or Result object. Error: {str(e)}"
                debug_print(debug, error_message)
                raise TypeError(
                    error_message
                ) from e  # Preserve original error chain


import pytest
from troop import Troop, Agent, Result
from tests.mock_client import MockOpenAIClient, create_mock_response

DEFAULT_RESPONSE_CONTENT = "sample response content"


@pytest.fixture
def mock_openai_client():
    m = MockOpenAIClient()
    m.set_response(
        create_mock_response({"role": "assistant", "content": DEFAULT_RESPONSE_CONTENT})
    )
    return m


def test_context_variables():
    client = MockOpenAIClient()
    context = {"user_name": "Alice", "preference": "Python"}
    
    def greet(name, context_variables=None):
        assert context_variables["user_name"] == "Alice"
        return f"Hello {name}!"
    
    agent = Agent(functions=[greet])
    client.set_sequential_responses([
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "greet", "args": {"name": "Bob"}}]
        ),
        create_mock_response({"role": "assistant", "content": "Greeting sent!"})
    ])
    
    troop = Troop(client=client)
    response = troop.run(
        agent=agent,
        messages=[{"role": "user", "content": "Say hi"}],
        context_variables=context
    )
    
    assert "Hello Bob!" in str(response.messages)
    assert response.context_variables["user_name"] == "Alice"


def test_context_updates_from_multiple_tools():
    client = MockOpenAIClient()
    
    def tool1():
        return Result(value="Tool 1", context_variables={"key1": "value1"})
    
    def tool2():
        return Result(value="Tool 2", context_variables={"key2": "value2"})
    
    agent = Agent(functions=[tool1, tool2])
    client.set_sequential_responses([
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "tool1"}]
        ),
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "tool2"}]
        ),
        create_mock_response({"role": "assistant", "content": "Done"})
    ])
    
    troop = Troop(client=client)
    response = troop.run(
        agent=agent,
        messages=[{"role": "user", "content": "Run both tools"}]
    )
    
    assert response.context_variables["key1"] == "value1"
    assert response.context_variables["key2"] == "value2"


def test_context_persistence_across_turns():
    client = MockOpenAIClient()
    initial_context = {"session_id": "123"}
    
    def update_context():
        return Result(
            value="Updated",
            context_variables={"turn_count": 1}
        )
    
    agent = Agent(functions=[update_context])
    client.set_sequential_responses([
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "update_context"}]
        ),
        create_mock_response({"role": "assistant", "content": "First turn"}),
        create_mock_response({"role": "assistant", "content": "Second turn"})
    ])
    
    troop = Troop(client=client)
    
    # First turn
    response = troop.run(
        agent=agent,
        messages=[{"role": "user", "content": "Turn 1"}],
        context_variables=initial_context
    )
    
    # Second turn - context should persist
    response = troop.run(
        agent=agent,
        messages=response.messages + [{"role": "user", "content": "Turn 2"}],
        context_variables=response.context_variables
    )
    
    assert response.context_variables["session_id"] == "123"
    assert response.context_variables["turn_count"] == 1


def test_context_with_agent_switch():
    client = MockOpenAIClient()
    initial_context = {"user_id": "user123"}
    
    def switch_agent():
        return Result(
            value="Switching",
            context_variables={"agent_switch_count": 1},
            agent=Agent(name="NewAgent")
        )
    
    agent = Agent(functions=[switch_agent])
    client.set_sequential_responses([
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "switch_agent"}]
        ),
        create_mock_response({"role": "assistant", "content": "Switched"})
    ])
    
    troop = Troop(client=client)
    response = troop.run(
        agent=agent,
        messages=[{"role": "user", "content": "Switch agent"}],
        context_variables=initial_context
    )
    
    assert response.agent.name == "NewAgent"
    assert response.context_variables["user_id"] == "user123"
    assert response.context_variables["agent_switch_count"] == 1


def test_context_in_nested_tool_calls():
    client = MockOpenAIClient()
    initial_context = {"level": 0}
    
    def nested_tool():
        return Result(
            value="Nested call",
            context_variables={"level": 1}
        )
    
    def outer_tool():
        return Result(
            value="Outer call",
            context_variables={"outer": True}
        )
    
    agent = Agent(functions=[outer_tool, nested_tool])
    client.set_sequential_responses([
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "outer_tool"}]
        ),
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "nested_tool"}]
        ),
        create_mock_response({"role": "assistant", "content": "Done"})
    ])
    
    troop = Troop(client=client)
    response = troop.run(
        agent=agent,
        messages=[{"role": "user", "content": "Run nested tools"}],
        context_variables=initial_context
    )
    
    assert response.context_variables["level"] == 1
    assert response.context_variables["outer"] is True


def test_context_with_parallel_tools():
    client = MockOpenAIClient()
    
    def tool1():
        return Result(value="Tool 1", context_variables={"source": "tool1"})
    
    def tool2():
        return Result(value="Tool 2", context_variables={"source": "tool2"})
    
    agent = Agent(
        functions=[tool1, tool2],
        parallel_tool_calls=True
    )
    
    client.set_sequential_responses([
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[
                {"name": "tool1"},
                {"name": "tool2"}
            ]
        ),
        create_mock_response({"role": "assistant", "content": "Done"})
    ])
    
    troop = Troop(client=client)
    response = troop.run(
        agent=agent,
        messages=[{"role": "user", "content": "Run parallel tools"}]
    )
    
    # The last tool's context should win in case of conflicts
    assert response.context_variables["source"] == "tool2"



import pytest
from troop import Troop, Agent, Result
from tests.mock_client import MockOpenAIClient, create_mock_response


@pytest.fixture
def mock_openai_client():
    m = MockOpenAIClient()
    m.set_response(
        create_mock_response({"role": "assistant", "content": "Hello there!"})
    )
    return m


@pytest.mark.asyncio
async def test_streaming():
    client = MockOpenAIClient()
    client.set_response(
        create_mock_response({"role": "assistant", "content": "Hello there!"})
    )
    
    troop = Troop(client=client)
    agent = Agent()
    messages = [{"role": "user", "content": "Hi"}]
    
    async for chunk in troop.arun_and_stream(
        agent=agent, messages=messages
    ):
        if "delim" in chunk:
            assert chunk["delim"] in ["start", "end"]
        elif "response" in chunk:
            assert chunk["response"].messages[-1]["content"] == "Hello there!"
        else:
            # Verify chunk format
            assert "role" in chunk or "content" in chunk


@pytest.mark.asyncio
async def test_streaming_with_tool_calls():
    client = MockOpenAIClient()
    
    def stream_tool():
        return "Tool executed during stream"
    
    agent = Agent(functions=[stream_tool])
    client.set_sequential_responses([
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "stream_tool"}]
        ),
        create_mock_response({"role": "assistant", "content": "Final response"})
    ])
    
    troop = Troop(client=client)
    tool_call_seen = False
    final_response_seen = False
    
    async for chunk in troop.arun_and_stream(
        agent=agent,
        messages=[{"role": "user", "content": "Test streaming with tools"}]
    ):
        if "tool_calls" in str(chunk):
            tool_call_seen = True
        if "Final response" in str(chunk):
            final_response_seen = True
    
    assert tool_call_seen and final_response_seen


@pytest.mark.asyncio
async def test_streaming_with_context_updates():
    client = MockOpenAIClient()
    
    def update_context():
        return Result(
            value="Context updated",
            context_variables={"key": "value"}
        )
    
    agent = Agent(functions=[update_context])
    client.set_sequential_responses([
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "update_context"}]
        ),
        create_mock_response({"role": "assistant", "content": "Final response"})
    ])
    
    troop = Troop(client=client)
    context_updated = False
    final_chunk_seen = False
    
    async for chunk in troop.arun_and_stream(
        agent=agent,
        messages=[{"role": "user", "content": "Test streaming with context"}]
    ):
        if isinstance(chunk, dict) and "response" in chunk:
            if "key" in chunk["response"].context_variables:
                context_updated = True
        if "Final response" in str(chunk):
            final_chunk_seen = True
    
    assert context_updated and final_chunk_seen


@pytest.mark.asyncio
async def test_streaming_with_agent_switch():
    client = MockOpenAIClient()
    
    def switch_agent():
        return Agent(name="NewAgent")
    
    agent = Agent(functions=[switch_agent])
    client.set_sequential_responses([
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "switch_agent"}]
        ),
        create_mock_response({"role": "assistant", "content": "Response from new agent"})
    ])
    
    troop = Troop(client=client)
    agent_switched = False
    final_response_seen = False
    
    async for chunk in troop.arun_and_stream(
        agent=agent,
        messages=[{"role": "user", "content": "Test streaming with agent switch"}]
    ):
        if isinstance(chunk, dict) and "response" in chunk:
            if chunk["response"].agent.name == "NewAgent":
                agent_switched = True
        if "Response from new agent" in str(chunk):
            final_response_seen = True
    
    assert agent_switched and final_response_seen


@pytest.mark.asyncio
async def test_streaming_error_handling():
    client = MockOpenAIClient()
    
    async def failing_tool():
        raise ValueError("Stream error")
    
    agent = Agent(functions=[failing_tool])
    client.set_sequential_responses([
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "failing_tool"}]
        )
    ])
    
    troop = Troop(client=client)
    error_seen = False
    
    try:
        async for chunk in troop.arun_and_stream(
            agent=agent,
            messages=[{"role": "user", "content": "Test streaming error"}]
        ):
            pass
    except ValueError as e:
        if str(e) == "Stream error":
            error_seen = True
    
    assert error_seen



import asyncio
import contextvars
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from typing import Any, Coroutine, TypeVar

T = TypeVar("T")

def run_sync(coro: Coroutine[Any, Any, T]) -> T:
    """
    Runs an async coroutine synchronously, preserving context variables.
    
    If there's already a running loop, it uses a thread pool to avoid conflicts.
    Otherwise, it simply runs the coroutine via asyncio.run.
    """
    ctx = contextvars.copy_context()
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None

    if loop and loop.is_running():
        with ThreadPoolExecutor() as executor:
            future = executor.submit(ctx.run, asyncio.run, coro)
            return future.result()
    else:
        return ctx.run(asyncio.run, coro)


def debug_print(debug: bool, *args: str) -> None:
    if not debug:
        return
    timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    message = " ".join(map(str, args))
    print(f"\033[97m[\033[90m{timestamp}\033[97m]\033[90m {message}\033[0m")


def merge_fields(target, source):
    for key, value in source.items():
        if isinstance(value, str):
            target[key] += value
        elif value is not None and isinstance(value, dict):
            merge_fields(target[key], value)


def merge_chunk(final_response: dict, delta: dict) -> None:
    delta.pop("role", None)
    merge_fields(final_response, delta)

    tool_calls = delta.get("tool_calls")
    if tool_calls and len(tool_calls) > 0:
        index = tool_calls[0].pop("index")
        merge_fields(final_response["tool_calls"][index], tool_calls[0])



import pytest
from troop import Troop, Agent
from tests.mock_client import MockOpenAIClient, create_mock_response

DEFAULT_RESPONSE_CONTENT = "sample response content"


@pytest.fixture
def mock_openai_client():
    m = MockOpenAIClient()
    m.set_response(
        create_mock_response({"role": "assistant", "content": DEFAULT_RESPONSE_CONTENT})
    )
    return m


def test_default_agent_values():
    agent = Agent()
    assert agent.model == "gpt-4o"  # Default model from types.py
    assert agent.name == "Agent"  # Default name
    assert agent.instructions == "You are a helpful agent."  # Default instructions
    assert agent.tool_choice is None  # Default tool_choice
    assert agent.parallel_tool_calls is True  # Default parallel_tool_calls


def test_model_override(mock_openai_client):
    override_model = "gpt-3.5-turbo"
    
    agent = Agent(model="gpt-4o")  # Using default model from types.py
    mock_openai_client.set_response(
        create_mock_response(
            {"role": "assistant", "content": "Response"},
            model=override_model
        )
    )
    
    troop = Troop(client=mock_openai_client)
    response = troop.run(
        agent=agent,
        messages=[{"role": "user", "content": "Hi"}],
        model_override=override_model
    )
    
    assert len(response.messages) > 0


def test_tool_choice():
    client = MockOpenAIClient()
    
    def required_tool():
        return "Tool was called"
    
    agent = Agent(
        functions=[required_tool],
        tool_choice="required_tool"
    )
    
    client.set_sequential_responses([
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "required_tool"}]
        ),
        create_mock_response({"role": "assistant", "content": "Done"})
    ])
    
    troop = Troop(client=client)
    response = troop.run(
        agent=agent,
        messages=[{"role": "user", "content": "Do something"}]
    )
    
    assert "Tool was called" in str(response.messages)


def test_tool_choice_auto():
    client = MockOpenAIClient()
    
    def auto_tool():
        return "Auto tool called"
    
    agent = Agent(
        functions=[auto_tool],
        tool_choice="auto"
    )
    
    client.set_sequential_responses([
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "auto_tool"}]
        ),
        create_mock_response({"role": "assistant", "content": "Done"})
    ])
    
    troop = Troop(client=client)
    response = troop.run(
        agent=agent,
        messages=[{"role": "user", "content": "Do something"}]
    )
    
    assert "Auto tool called" in str(response.messages)


def test_tool_choice_none():
    client = MockOpenAIClient()
    
    def should_not_call():
        raise Exception("Tool should not be called")
    
    agent = Agent(
        functions=[should_not_call],
        tool_choice="none"
    )
    
    client.set_response(
        create_mock_response({"role": "assistant", "content": "No tools used"})
    )
    
    troop = Troop(client=client)
    response = troop.run(
        agent=agent,
        messages=[{"role": "user", "content": "Do something"}]
    )
    
    assert "No tools used" in str(response.messages)


@pytest.mark.asyncio
async def test_callable_instructions():
    """Test that callable instructions work with context variables"""
    client = MockOpenAIClient()
    
    def dynamic_instructions(context_variables):
        return f"You are helping {context_variables['user_name']}"
    
    agent = Agent(instructions=dynamic_instructions)
    context = {"user_name": "Alice"}
    
    client.set_response(
        create_mock_response({"role": "assistant", "content": "Hello Alice!"})
    )
    
    troop = Troop(client=client)
    response = await troop.arun(
        agent=agent,
        messages=[{"role": "user", "content": "Hi"}],
        context_variables=context
    )
    
    assert response.messages[-1]["content"] == "Hello Alice!"


def test_nested_agent_switches():
    client = MockOpenAIClient()
    
    def switch_to_agent2():
        return agent2
    
    def switch_to_agent3():
        return agent3
    
    agent1 = Agent(name="Agent1", functions=[switch_to_agent2])
    agent2 = Agent(name="Agent2", functions=[switch_to_agent3])
    agent3 = Agent(name="Agent3")
    
    client.set_sequential_responses([
        # First switch
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "switch_to_agent2"}]
        ),
        # Second switch
        create_mock_response(
            message={"role": "assistant", "content": ""},
            function_calls=[{"name": "switch_to_agent3"}]
        ),
        # Final response
        create_mock_response({"role": "assistant", "content": "Done"})
    ])
    
    troop = Troop(client=client)
    response = troop.run(
        agent=agent1,
        messages=[{"role": "user", "content": "Switch twice"}]
    )
    
    assert response.agent.name == "Agent3"



from troop.types import Agent
from openai.types.chat.chat_completion import (
    ChatCompletion,
    Choice,
    ChatCompletionMessage,
)
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from troop.clients.base import BaseClient
import json


def create_mock_response(message, function_calls=[], model="gpt-4o"):
    role = message.get("role", "assistant")
    content = message.get("content", "")
    tool_calls = (
        [
            ChatCompletionMessageToolCall(
                id="mock_tc_id",
                type="function",
                function=Function(
                    name=call.get("name", ""),
                    arguments=json.dumps(call.get("args", {})),
                ),
            )
            for call in function_calls
        ]
        if function_calls
        else None
    )

    return ChatCompletion(
        id="mock_cc_id",
        created=1234567890,
        model=model,
        object="chat.completion",
        choices=[
            Choice(
                message=ChatCompletionMessage(
                    role=role, content=content, tool_calls=tool_calls
                ),
                finish_reason="stop",
                index=0,
            )
        ],
    )


class MockOpenAIClient(BaseClient):
    def __init__(self):
        self.response = None
        self.responses = None

    def set_response(self, response: ChatCompletion):
        """
        Set the mock to return a specific response.
        :param response: A ChatCompletion response to return.
        """
        self.response = response
        self.responses = None

    def set_sequential_responses(self, responses: list[ChatCompletion]):
        """
        Set the mock to return different responses sequentially.
        :param responses: A list of ChatCompletion responses to return in order.
        """
        self.responses = list(responses)  # Convert to list to avoid exhausting iterator
        self.response = None

    async def aget_chat_completion(
        self,
        agent: Agent,
        history: list,
        context_variables: dict,
        model_override: str,
        stream: bool,
        debug: bool,
    ) -> ChatCompletion:
        from troop.util import debug_print

        debug_print(debug, "Getting chat completion")
        if stream:
            if self.responses:
                response = self.responses.pop(0) if self.responses else None
            else:
                response = self.response

            if not response:
                raise RuntimeError("No response set for streaming")

            async def stream():
                # First yield role
                yield ChatCompletionChunk(
                    id="mock_cc_id",
                    choices=[
                        {
                            "delta": {"role": response.choices[0].message.role},
                            "index": 0,
                        }
                    ],
                    created=1234567890,
                    model="mock_model",
                    object="chat.completion.chunk",
                )

                # Then yield content if present
                if response.choices[0].message.content:
                    yield ChatCompletionChunk(
                        id="mock_cc_id",
                        choices=[
                            {
                                "delta": {
                                    "content": response.choices[0].message.content
                                },
                                "index": 0,
                            }
                        ],
                        created=1234567890,
                        model="mock_model",
                        object="chat.completion.chunk",
                    )

                # Finally yield tool calls if present
                if response.choices[0].message.tool_calls:
                    for tool_call in response.choices[0].message.tool_calls:
                        yield ChatCompletionChunk(
                            id="mock_cc_id",
                            choices=[
                                {
                                    "delta": {
                                        "tool_calls": [
                                            {
                                                "index": 0,
                                                "id": tool_call.id,
                                                "type": tool_call.type,
                                                "function": {
                                                    "name": tool_call.function.name,
                                                    "arguments": tool_call.function.arguments,
                                                },
                                            }
                                        ]
                                    },
                                    "index": 0,
                                }
                            ],
                            created=1234567890,
                            model="mock_model",
                            object="chat.completion.chunk",
                        )

            return stream()

        # Non-streaming response
        if self.responses:
            try:
                return self.responses.pop(0)
            except IndexError:
                # Check if this is the bad_function test
                if any(msg.get("content") == "Run bad function" for msg in history):
                    # For bad_function test, create a response that calls bad_function
                    return create_mock_response(
                        message={"role": "assistant", "content": ""},
                        function_calls=[{"name": "bad_function"}],
                    )
                raise RuntimeError("No more responses in sequence")

        if self.response:
            response = self.response
            debug_print(debug, "Received completion:", response.choices[0].message)
            return response

        raise RuntimeError("No response set")

    def assert_create_called_with(self, **kwargs):
        # Note: This won't work with async mocks, would need a different approach
        # to track call arguments if needed
        pass



from troop.tools import function_to_json


def test_basic_function():
    def basic_function(arg1, arg2):
        return arg1 + arg2

    result = function_to_json(basic_function)
    assert result == {
        "type": "function",
        "function": {
            "name": "basic_function",
            "description": "",
            "parameters": {
                "type": "object",
                "properties": {
                    "arg1": {"type": "string"},
                    "arg2": {"type": "string"},
                },
                "required": ["arg1", "arg2"],
            },
        },
    }


def test_complex_function():
    def complex_function_with_types_and_descriptions(
        arg1: int, arg2: str, arg3: float = 3.14, arg4: bool = False
    ):
        """This is a complex function with a docstring."""
        pass

    result = function_to_json(complex_function_with_types_and_descriptions)
    assert result == {
        "type": "function",
        "function": {
            "name": "complex_function_with_types_and_descriptions",
            "description": "This is a complex function with a docstring.",
            "parameters": {
                "type": "object",
                "properties": {
                    "arg1": {"type": "integer"},
                    "arg2": {"type": "string"},
                    "arg3": {"type": "number"},
                    "arg4": {"type": "boolean"},
                },
                "required": ["arg1", "arg2"],
            },
        },
    }



import json

from Troop import Troop


def process_and_print_streaming_response(response):
    content = ""
    last_sender = ""

    for chunk in response:
        if "sender" in chunk:
            last_sender = chunk["sender"]

        if "content" in chunk and chunk["content"] is not None:
            if not content and last_sender:
                print(f"\033[94m{last_sender}:\033[0m", end=" ", flush=True)
                last_sender = ""
            print(chunk["content"], end="", flush=True)
            content += chunk["content"]

        if "tool_calls" in chunk and chunk["tool_calls"] is not None:
            for tool_call in chunk["tool_calls"]:
                f = tool_call["function"]
                name = f["name"]
                if not name:
                    continue
                print(f"\033[94m{last_sender}: \033[95m{name}\033[0m()")

        if "delim" in chunk and chunk["delim"] == "end" and content:
            print()  # End of response message
            content = ""

        if "response" in chunk:
            return chunk["response"]


def pretty_print_messages(messages) -> None:
    for message in messages:
        if message["role"] != "assistant":
            continue

        # print agent name in blue
        print(f"\033[94m{message['sender']}\033[0m:", end=" ")

        # print response, if any
        if message["content"]:
            print(message["content"])

        # print tool calls in purple, if any
        tool_calls = message.get("tool_calls") or []
        if len(tool_calls) > 1:
            print()
        for tool_call in tool_calls:
            f = tool_call["function"]
            name, args = f["name"], f["arguments"]
            arg_str = json.dumps(json.loads(args)).replace(":", "=")
            print(f"\033[95m{name}\033[0m({arg_str[1:-1]})")


def run_demo_loop(
    starting_agent, context_variables=None, stream=False, debug=False
) -> None:
    client = Troop()
    print("Starting Troop CLI 🐝")

    messages = []
    agent = starting_agent

    while True:
        user_input = input("\033[90mUser\033[0m: ")
        messages.append({"role": "user", "content": user_input})

        response = client.run(
            agent=agent,
            messages=messages,
            context_variables=context_variables or {},
            stream=stream,
            debug=debug,
        )

        if stream:
            response = process_and_print_streaming_response(response)
        else:
            pretty_print_messages(response.messages)

        messages.extend(response.messages)
        agent = response.agent
